@@ -1,8 +1,10 @@
 package PAR::Dist;
+use 5.006;
+use strict;
 require Exporter;
-use vars qw/$VERSION @ISA @EXPORT @EXPORT_OK/;
+use vars qw/$VERSION @ISA @EXPORT @EXPORT_OK $DEBUG/;
 
-$VERSION    = '0.31';
+$VERSION    = '0.45';
 @ISA	    = 'Exporter';
 @EXPORT	    = qw/
   blib_to_par
@@ -21,7 +23,8 @@ $VERSION    = '0.31';
   contains_binaries
 /;
 
-use strict;
+$DEBUG = 0;
+
 use Carp qw/carp croak/;
 use File::Spec;
 
@@ -31,7 +34,7 @@ PAR::Dist - Create and manipulate PAR distributions
 
 =head1 VERSION
 
-This document describes version 0.31 of PAR::Dist, released May 28, 2008.
+This document describes version 0.45 of PAR::Dist, released June 11, 2009.
 
 =head1 SYNOPSIS
 
@@ -105,7 +108,7 @@ under the current directory if unspecified.  If F<blib/> does not exist,
 it automatically runs F<Build>, F<make>, F<Build.PL> or F<Makefile.PL> to
 create it.
 
-Returns the filename or the generated PAR distribution.
+Returns the filename of the generated PAR distribution.
 
 Valid parameters are:
 
@@ -129,6 +132,10 @@ perl by default.
 
 The output filename for the PAR distribution.
 
+=item quiet
+
+Set to true to suppress as much output as possible.
+
 =back
 
 =cut
@@ -141,6 +148,7 @@ sub blib_to_par {
 
 
     # don't use 'my $foo ... if ...' it creates a static variable!
+    my $quiet = $args{quiet} || 0;
     my $dist;
     my $path	= $args{path};
     $dist	= File::Spec->rel2abs($args{dist}) if $args{dist};
@@ -250,7 +258,9 @@ YAML
 
     mkdir('blib', 0777);
     chdir('blib');
-    _zip(dist => File::Spec->catfile(File::Spec->updir, $file)) or die $!;
+    require Cwd;
+    my $zipoutfile = File::Spec->catfile(File::Spec->updir, $file);
+    _zip(dist => $zipoutfile);
     chdir(File::Spec->updir);
 
     unlink File::Spec->catfile("blib", "MANIFEST");
@@ -259,7 +269,12 @@ YAML
     $dist ||= File::Spec->catfile($cwd, $file) if $cwd;
 
     if ($dist and $file ne $dist) {
-        rename( $file => $dist );
+        if ( File::Copy::copy($file, $dist) ) {
+          unlink $file;
+        } else {
+          die "Cannot copy $file: $!";
+        }
+
         $file = $dist;
     }
 
@@ -268,7 +283,7 @@ YAML
         $pathname =~ s!\\!/!g;
         $pathname =~ s!:!|!g;
     };
-    print << ".";
+    print << "." if !$quiet;
 Successfully created binary distribution '$file'.
 Its contents are accessible in compliant browsers as:
     jar:file://$pathname!/MANIFEST
@@ -280,18 +295,18 @@ Its contents are accessible in compliant browsers as:
 
 sub _build_blib {
     if (-e 'Build') {
-        system($^X, "Build");
+        _system_wrapper($^X, "Build");
     }
     elsif (-e 'Makefile') {
-        system($Config::Config{make});
+        _system_wrapper($Config::Config{make});
     }
     elsif (-e 'Build.PL') {
-        system($^X, "Build.PL");
-        system($^X, "Build");
+        _system_wrapper($^X, "Build.PL");
+        _system_wrapper($^X, "Build");
     }
     elsif (-e 'Makefile.PL') {
-        system($^X, "Makefile.PL");
-        system($Config::Config{make});
+        _system_wrapper($^X, "Makefile.PL");
+        _system_wrapper($Config::Config{make});
     }
 }
 
@@ -300,7 +315,10 @@ sub _build_blib {
 Installs a PAR distribution into the system, using
 C<ExtUtils::Install::install_default>.
 
-Valid parameters are:
+If only a single parameter is given, it is treated as the value for the
+C<dist> parameter.
+
+Valid named parameters are:
 
 =over 2
 
@@ -315,8 +333,25 @@ This string will be prepended to all installation paths.
 If it isn't specified, the environment variable
 C<PERL_INSTALL_ROOT> is used as a prefix.
 
+=item uninstall_shadows
+
+This corresponds to the C<uninstall_shadows> option of L<ExtUtils::Install>. Quoting its manual:
+If C<uninstall_shadows> is set to true, any differing versions throughout C<@INC>
+will be uninstalled. This is C<make install UNINST=1>.
+
+=item verbose
+
+This corresponds to the C<verbose> option of L<ExtUtils::Install>. According to its manual:
+If C<verbose> is true, will print out each file removed. This is C<make install VERBINST=1>.
+C<verbose> values going up to 5 show increasingly more diagnostics output.
+
+Default verbosity for PAR::Dist is 1.
+
 =back
 
+If you're just going to install into the running perl like everything else,
+you can stop reading the rest of this section now.
+
 Additionally, you can use several parameters to change the default
 installation destinations. You don't usually have to worry about this
 unless you are installing into a user-local directory.
@@ -340,6 +375,16 @@ you can remove that target altogether. For example, passing
 C<inst_man1dir => undef, inst_man3dir => undef> means that the contained
 manual pages won't be installed. This is not available for the packlists.
 
+Again, the defaults will be the normal I<site> paths from C<%Config>.
+
+(*) If the C<.par>'s I<inst_archlib> section (normally C<blib/arch>)
+isn't empty, the code in I<inst_lib> (normally C<blib/lib>) is also installed
+into the I<inst_archlib> path. This makes sense for XS modules.
+If, however, you override C<inst_lib>, this automatic conversion is
+also overridden! You can use the named parameter
+C<auto_inst_lib_conversion =E<gt> 1> to re-enable the conversion
+for custom I<inst_lib>'s.
+
 Finally, you may specify a C<custom_targets> parameter. Its value should be
 a reference to a hash of custom installation targets such as
 
@@ -348,9 +393,6 @@ a reference to a hash of custom installation targets such as
 You can use this to install the F<.par> archives contents to arbitrary
 locations.
 
-If only a single parameter is given, it is treated as the C<dist>
-parameter.
-
 =cut
 
 sub install_par {
@@ -369,6 +411,9 @@ to this is the C<packlist_read> parameter which specifies the
 F<.packlist> file to read the list of installed files from.
 It defaults to C<$Config::Config{installsitearch}/auto/$name/.packlist>.
 
+Additionally, the C<uninstall_shadows> parameter of C<install_par>
+isn't available.
+
 =cut
 
 sub uninstall_par {
@@ -412,18 +457,21 @@ sub _install_or_uninstall {
     $name =~ s{::|-}{/}g;
     require ExtUtils::Install;
 
-    my $rv;
     if ($action eq 'install') {
         my $target = _installation_target( File::Spec->curdir, $name, \%args );
         my $custom_targets = $args{custom_targets} || {};
         $target->{$_} = $custom_targets->{$_} foreach keys %{$custom_targets};
         
-        $rv = ExtUtils::Install::install($target, 1, 0, 0);
+        my $uninstall_shadows = $args{uninstall_shadows};
+        my $verbose = $args{verbose};
+        ExtUtils::Install::install($target, $verbose, 0, $uninstall_shadows);
     }
     elsif ($action eq 'uninstall') {
         require Config;
-        $rv = ExtUtils::Install::uninstall(
-            $args{packlist_read}||"$Config::Config{installsitearch}/auto/$name/.packlist"
+        my $verbose = $args{verbose};
+        ExtUtils::Install::uninstall(
+            $args{packlist_read}||"$Config::Config{installsitearch}/auto/$name/.packlist",
+            $verbose
         );
     }
 
@@ -431,7 +479,8 @@ sub _install_or_uninstall {
 
     chdir($old_dir);
     File::Path::rmtree([$tmpdir]);
-    return $rv;
+
+    return 1;
 }
 
 # Returns the default installation target as used by
@@ -460,14 +509,16 @@ sub _installation_target {
     );
 
 
+    my $par_has_archlib = _directory_not_empty( $sources{inst_archlib} );
+
     # default targets
     my $target = {
        read => $Config::Config{sitearchexp}."/auto/$name/.packlist",
        write => $Config::Config{installsitearch}."/auto/$name/.packlist",
-       $sources{inst_lib}
-            => (_directory_not_empty($sources{inst_archlib}))
-            ? $Config::Config{installsitearch}
-            : $Config::Config{installsitelib},
+       $sources{inst_lib} =>
+            ($par_has_archlib
+             ? $Config::Config{installsitearch}
+             : $Config::Config{installsitelib}),
        $sources{inst_archlib}   => $Config::Config{installsitearch},
        $sources{inst_bin}       => $Config::Config{installbin} ,
        $sources{inst_script}    => $Config::Config{installscript},
@@ -502,6 +553,12 @@ sub _installation_target {
         }
     }
 
+    # apply the automatic inst_lib => inst_archlib conversion again
+    # if the user asks for it and there is an archlib in the .par
+    if ($user->{auto_inst_lib_conversion} and $par_has_archlib) {
+      $target->{inst_lib} = $target->{inst_archlib};
+    }
+
     return $target;
 }
 
@@ -549,7 +606,12 @@ sub verify_par {
 
 =head2 merge_par
 
-Merge two or more PAR distributions into one. First argument must
+I<Note:> Since version 0.32 of PAR::Dist, this function requires a YAML
+reader. The order of precedence is:
+
+  YAML YAML::Syck YAML::Tiny YAML::XS
+
+Merges two or more PAR distributions into one. First argument must
 be the name of the distribution you want to merge all others into.
 Any following arguments will be interpreted as the file names of
 further PAR distributions to merge into the first one.
@@ -558,7 +620,10 @@ further PAR distributions to merge into the first one.
 
 This will merge the distributions C<foo.par>, C<bar.par> and C<baz.par>
 into the distribution C<foo.par>. C<foo.par> will be overwritten!
-The original META.yml of C<foo.par> is retained.
+
+The original META.yml of C<foo.par> is retained, but augmented with any
+C<provides>, C<requires>, C<recommends>, C<build_requires>, and
+C<configure_requires> sections from the other C<.par> files.
 
 =cut
 
@@ -595,9 +660,10 @@ sub merge_par {
     my $blibdir = File::Spec->catdir($base_dir, 'blib');
 
     # move the META.yml to the (main) temp. dir.
+    my $main_meta_file = File::Spec->catfile($base_dir, 'META.yml');
     File::Copy::move(
         File::Spec->catfile($blibdir, 'META.yml'),
-        File::Spec->catfile($base_dir, 'META.yml')
+        $main_meta_file
     );
     # delete (incorrect) MANIFEST
     unlink File::Spec->catfile($blibdir, 'MANIFEST');
@@ -610,6 +676,13 @@ sub merge_par {
         (undef, my $add_dir) = _unzip_to_tmpdir(
             dist => $par
         );
+
+        # merge the meta (at least the provides info) into the main meta.yml
+        my $meta_file = File::Spec->catfile($add_dir, 'META.yml');
+        if (-f $meta_file) {
+          _merge_meta($main_meta_file, $meta_file);
+        }
+
         my @files;
         my @dirs;
         # I hate File::Find
@@ -656,7 +729,7 @@ sub merge_par {
     unlink File::Spec->catfile($blibdir, 'META.yml');
     
     chdir($base_dir);
-    my $resulting_par_file = Cwd::abs_path(blib_to_par());
+    my $resulting_par_file = Cwd::abs_path(blib_to_par(quiet => 1));
     chdir($old_cwd);
     File::Copy::move($resulting_par_file, $base_par);
     
@@ -664,6 +737,72 @@ sub merge_par {
 }
 
 
+sub _merge_meta {
+  my $meta_orig_file = shift;
+  my $meta_extra_file = shift;
+  return() if not defined $meta_orig_file or not -f $meta_orig_file;
+  return 1 if not defined $meta_extra_file or not -f $meta_extra_file;
+
+  my $yaml_functions = _get_yaml_functions();
+
+  die "Cannot merge META.yml files without a YAML reader/writer"
+    if !exists $yaml_functions->{LoadFile}
+    or !exists $yaml_functions->{DumpFile};
+
+  my $orig_meta  = $yaml_functions->{LoadFile}->($meta_orig_file);
+  my $extra_meta = $yaml_functions->{LoadFile}->($meta_extra_file);
+
+  # I seem to remember there was this incompatibility between the different
+  # YAML implementations with regards to "document" handling:
+  my $orig_tree  = (ref($orig_meta) eq 'ARRAY' ? $orig_meta->[0] : $orig_meta);
+  my $extra_tree = (ref($extra_meta) eq 'ARRAY' ? $extra_meta->[0] : $extra_meta);
+
+  _merge_provides($orig_tree, $extra_tree);
+  _merge_requires($orig_tree, $extra_tree);
+  
+  $yaml_functions->{DumpFile}->($meta_orig_file, $orig_meta);
+
+  return 1;
+}
+
+# merge the two-level provides sections of META.yml
+sub _merge_provides {
+  my $orig_hash  = shift;
+  my $extra_hash = shift;
+
+  return() if not exists $extra_hash->{provides};
+  $orig_hash->{provides} ||= {};
+
+  my $orig_provides  = $orig_hash->{provides};
+  my $extra_provides = $extra_hash->{provides};
+
+  # two level clone is enough wrt META spec 1.4
+  # overwrite the original provides since we're also overwriting the files.
+  foreach my $module (keys %$extra_provides) {
+    my $extra_mod_hash = $extra_provides->{$module};
+    my %mod_hash;
+    $mod_hash{$_} = $extra_mod_hash->{$_} for keys %$extra_mod_hash;
+    $orig_provides->{$module} = \%mod_hash;
+  }
+}
+
+# merge the single-level requires-like sections of META.yml
+sub _merge_requires {
+  my $orig_hash  = shift;
+  my $extra_hash = shift;
+
+  foreach my $type (qw(requires build_requires configure_requires recommends)) {
+    next if not exists $extra_hash->{$type};
+    $orig_hash->{$type} ||= {};
+    
+    # one level clone is enough wrt META spec 1.4
+    foreach my $module (keys %{ $extra_hash->{$type} }) {
+      # FIXME there should be a version comparison here, BUT how are we going to do that without a guaranteed version.pm?
+      $orig_hash->{$type}{$module} = $extra_hash->{$type}{$module}; # assign version and module name
+    }
+  }
+}
+
 =head2 remove_man
 
 Remove the man pages from a PAR distribution. Takes one named
@@ -802,7 +941,11 @@ sub _unzip {
     }
     # Then fall back to the system
     else {
-        return if system(unzip => $dist, '-d', $path);
+        undef $!;
+        if (_system_wrapper(unzip => $dist, '-d', $path)) {
+            die "Failed to unzip '$dist' to path '$path': Could neither load "
+                . "Archive::Zip nor (successfully) run the system 'unzip' (unzip said: $!)";
+        }
     }
 
     return 1;
@@ -818,8 +961,13 @@ sub _zip {
         $zip->writeToFileNamed( $dist ) == Archive::Zip::AZ_OK() or die $!;
     }
     else {
-        system(qw(zip -r), $dist, File::Spec->curdir) and die $!;
+        undef $!;
+        if (_system_wrapper(qw(zip -r), $dist, File::Spec->curdir)) {
+            die "Failed to zip '" .File::Spec->curdir(). "' to '$dist': Could neither load "
+                . "Archive::Zip nor (successfully) run the system 'zip' (zip said: $!)";
+        }
     }
+    return 1;
 }
 
 
@@ -1161,6 +1309,91 @@ sub contains_binaries {
     return $found ? 1 : 0;
 }
 
+sub _system_wrapper {
+  if ($DEBUG) {
+    Carp::cluck("Running system call '@_' from:");
+  }
+  return system(@_);
+}
+
+# stolen from Module::Install::Can
+# very much internal and subject to change or removal
+sub _MI_can_run {
+  require ExtUtils::MakeMaker;
+  my ($cmd) = @_;
+
+  my $_cmd = $cmd;
+  return $_cmd if (-x $_cmd or $_cmd = MM->maybe_command($_cmd));
+
+  for my $dir ((split /$Config::Config{path_sep}/, $ENV{PATH}), '.') {
+    my $abs = File::Spec->catfile($dir, $cmd);
+    return $abs if (-x $abs or $abs = MM->maybe_command($abs));
+  }
+
+  return;
+}
+
+
+# Tries to load any YAML reader writer I know of
+# returns nothing on failure or hash reference containing
+# a subset of Load, Dump, LoadFile, DumpFile
+# entries with sub references on success.
+sub _get_yaml_functions {
+  # reasoning for the ranking here:
+  # - syck is fast and reasonably complete
+  # - YAML.pm is slow and aging
+  # - Tiny is only a very small subset
+  # - XS is very new and I'm not sure it's ready for prime-time yet
+  # - Parse... is only a reader and only deals with the same subset as ::Tiny
+  my @modules = qw(YAML::Syck YAML YAML::Tiny YAML::XS Parse::CPAN::Meta);
+
+  my %yaml_functions;
+  foreach my $module (@modules) {
+    eval "require $module;";
+    if (!$@) {
+      warn "PAR::Dist testers/debug info: Using '$module' as YAML implementation" if $DEBUG;
+      foreach my $sub (qw(Load Dump LoadFile DumpFile)) {
+        no strict 'refs';
+        my $subref = *{"${module}::$sub"}{CODE};
+        if (defined $subref and ref($subref) eq 'CODE') {
+          $yaml_functions{$sub} = $subref;
+        }
+      }
+      $yaml_functions{yaml_provider} = $module;
+      last;
+    }
+  } # end foreach module candidates
+  if (not keys %yaml_functions) {
+    warn "Cannot find a working YAML reader/writer implementation. Tried to load all of '@modules'";
+  }
+  return(\%yaml_functions);
+}
+
+sub _check_tools {
+  my $tools = _get_yaml_functions();
+  if ($DEBUG) {
+    foreach (qw/Load Dump LoadFile DumpFile/) {
+      warn "No YAML support for $_ found.\n" if not defined $tools->{$_};
+    }
+  }
+
+  $tools->{zip} = undef;
+  if (eval {require Archive::Zip; 1;}) {
+    warn "Using Archive::Zip as ZIP tool.\n" if $DEBUG;
+    $tools->{zip} = 'Archive::Zip';
+  }
+  elsif (_MI_can_run("zip") and _MI_can_run("unzip")) {
+    warn "Using zip/unzip as ZIP tool.\n" if $DEBUG;
+    $tools->{zip} = 'zip';
+  }
+  else {
+    warn "Found neither Archive::Zip nor ZIP/UNZIP as valid ZIP tools.\n" if $DEBUG;
+    $tools->{zip} = undef;
+  }
+
+  return $tools;
+}
+
 1;
 
 =head1 SEE ALSO
@@ -1171,7 +1404,7 @@ L<PAR>, L<ExtUtils::Install>, L<Module::Signature>, L<LWP::Simple>
 
 Audrey Tang E<lt>cpan@audreyt.orgE<gt> 2003-2007
 
-Steffen Mueller E<lt>smueller@cpan.orgE<gt> 2005-2007
+Steffen Mueller E<lt>smueller@cpan.orgE<gt> 2005-2008
 
 PAR has a mailing list, E<lt>par@perl.orgE<gt>, that you can write to;
 send an empty mail to E<lt>par-subscribe@perl.orgE<gt> to join the list
@@ -1181,7 +1414,7 @@ Please send bug reports to E<lt>bug-par@rt.cpan.orgE<gt>.
 
 =head1 COPYRIGHT
 
-Copyright 2003-2007 by Audrey Tang E<lt>autrijus@autrijus.orgE<gt>.
+Copyright 2003-2008 by Audrey Tang E<lt>autrijus@autrijus.orgE<gt>.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
